
open Format

open Utils
open Types
open Lattice
open Expr.StdExpression
open FixpointAnalysis

(* stat is defined outer, to ensure that all classical reachibility analyses have the same stat type *)
module Graph = LabeledTransitionSystem.Graph
type stat = Fixpoint.stat


module type ClassicalReachabilityAnalysis =
sig
  module Configuration :
  sig
    val argument_specification : Configuration.t -> (Arg.key * Arg.spec * Arg.doc) list
    val default_configuration : Configuration.values
  end

  type ('s,'t,'p) input = ('s, 't*guard*(command list), 'p) LabeledTransitionSystem.t

  val run_and_print :
    Utils.Configuration.t ->
    ('s,'t,'p) input ->
    's printer ->
    't printer ->
    'p printer ->
    Format.formatter ->
    unit
end


(* TODO : allow manager *)

module Reachability (Lattice : ManagedSemiLattice with type manager = unit) : ClassicalReachabilityAnalysis =
struct
  module Configuration =
  struct
    let set_widening_descend configuration step =
      Configuration.set_int configuration "descendings" (max 0 step)

    let set_widening configuration step =
      Configuration.set_int configuration "widening" (max 0 step)

    let argument_specification configuration =
      [("-widening",
        Arg.Int (set_widening configuration),
        "n  Set n step before applying widening (negative or null means no widening)") ;
       ("-descending",
        Arg.Int (set_widening_descend configuration),
         "n  Set n step of descendings after widening (negative or null means no descending)") ;
       ("-dot",
        Arg.Unit (Configuration.set_flag configuration "dot_output"),
        " Print results as a dot graph in the standard output") ;
       ("-stats",
        Arg.Unit (Configuration.set_flag configuration "stats"),
        "   Print statistics of analysis")
      ]

    let default_configuration =
    [
      ("descendings", Configuration.Int 0) ;
      ("widening", Configuration.Int 1) ;
      ("stats", Configuration.Bool false)
    ]
  end

  (* types of arc and abstract used by MkFixpoint *)

  type vertex = Graph.vertex
  type hedge = Graph.hedge
  type abstract = Lattice.t
  type arc = unit
  type fixpoint_manager = (vertex, hedge, abstract, arc) Fixpoint.manager


  (* Types of the input and output of the analysis *)
  type l_guard = Lattice.Semantics.Guard.t
  type l_command = Lattice.Semantics.Command.t

  type ('s,'t,'p) input_graph = ('s, 't*guard*(command list), 'p) LabeledTransitionSystem.Graph.t
  type ('s,'t,'p) converted_graph = ('s, 't*l_guard*(l_command list), 'p) LabeledTransitionSystem.Graph.t
  type output_graph = (vertex, hedge, abstract, arc) Fixpoint.output
  type ('s,'t,'p) input = ('s, 't*guard*(command list), 'p) LabeledTransitionSystem.t


  (* Analysis *)

  (* Conversion of guard and commands to Lattice expressions *)
  let convert_graph (program_graph : ('s,'t,'p) input_graph) (lattice_manager : Lattice.manager) : ('s,'t,'p) converted_graph =
    let f_vertex (v : vertex) (s : 's) =
      s
    in
    let f_hedge (h : hedge) (t,guard,commands : 't*guard*(command list)) =
      (
        t,
        Lattice.Semantics.Guard.from_guard lattice_manager guard,
        List.map (Lattice.Semantics.Command.from_command lattice_manager) commands
      )
    in
    let f_info (p : 'p) =
      p
    in
      Graph.map program_graph f_vertex f_hedge f_info

  let init_lattice_manager () = ()

  let init_fixpoint_manager
    (configuration : Utils.Configuration.t)
    (converted_graph : ('s,'t,'p) Graph.t)
    (init_states : Graph.SetV.t)
    (lattice_manager : Lattice.manager)
  : fixpoint_manager
  =
    (* Initialisation *)
    let abstract_init v =
      if Graph.SetV.mem v init_states then
        Lattice.top lattice_manager
      else
        Lattice.bottom lattice_manager
    in
    (* Application of guard and commands to an abstract value *)
    let apply (e : hedge) (av_array : abstract array) : arc * abstract =
      let (_, guard, commands) = Graph.attrhedge converted_graph e in
      let av = av_array.(0) in
      let av_g, _ = Lattice.Semantics.apply_guard lattice_manager av guard in
      let av_c = List.fold_left
        (fun a c -> fst (Lattice.Semantics.apply_command lattice_manager a c))
        av_g
        commands
      in
        ((), av_c)
    (* Finally the manager *)
    in
    {
      (* Lattice operations. Vertice symbol v could be used *)
      Fixpoint.bottom = (fun v -> Lattice.bottom lattice_manager) ;
      canonical = (fun v -> Lattice.canonical lattice_manager) ;
      is_bottom = (fun v -> Lattice.is_bottom lattice_manager) ;
      is_leq = (fun v -> Lattice.is_leq lattice_manager) ;
      join = (fun v -> Lattice.join lattice_manager) ;
      join_list = (fun v -> Lattice.join_list lattice_manager) ;
      widening = (fun v -> Lattice.widening lattice_manager) ;
      (* Functions just defined *)
      apply = apply ;
      arc_init = (fun hedge -> ()) ;
      abstract_init = abstract_init ;
      (* Options *)
      accumulate = false ;
      widening_first = true ;
      widening_start = Utils.Configuration.get_int configuration "widening" ;
      widening_freq = 1 ;
      widening_descend = Utils.Configuration.get_int configuration "descendings" ;
      print_analysis = (Utils.Configuration.get_int configuration "verbose") > 0 ;
      print_component = (Utils.Configuration.get_int configuration "verbose") > 0 ;
      print_step = (Utils.Configuration.get_int configuration "verbose") > 0 ;
      print_state = (Utils.Configuration.get_int configuration "verbose") > 1 ;
      print_postpre = (Utils.Configuration.get_int configuration "verbose") > 2 ;
      print_workingsets = (Utils.Configuration.get_int configuration "verbose") > 3 ; (* Impossible *)
      (* Printing functions *)
      print_fmt = err_formatter ;
      print_abstract = Lattice.print lattice_manager ;
      print_arc = (fun fmt () -> ()) ;
      print_vertex = Symbol.print ;
      print_hedge = Symbol.print ;
      dot_fmt = None ;
      dot_vertex = Symbol.print ;
      dot_hedge = Symbol.print ;
      dot_attrvertex = Symbol.print ;
      dot_attrhedge = Symbol.print ;
    }


  (* Run the analysis *)
  let run_fixpoint_analysis
    (configuration : Utils.Configuration.t)
    (program_graph : ('s,'t,'p) input_graph)
    (init_states : Graph.SetV.t)
    (lattice_manager : Lattice.manager)
  :
    output_graph
  =
    (* Convert expression to Lattice expressions *)
    let converted_graph = convert_graph program_graph lattice_manager in
    (* Definition of a MkFixpoint.manager *)
    let fixpoint_manager = init_fixpoint_manager configuration converted_graph init_states lattice_manager in
    (* Define strategy *)
    let module T = LabeledTransitionSystem.TSymbol in
    let strategy = Fixpoint.make_strategy_default T.vertex_dummy T.hedge_dummy program_graph init_states in
    (* Compute the fixpoint *)
    Fixpoint.analysis_std fixpoint_manager converted_graph init_states strategy


  (* Output printing *)

  let print_stats fmt {Fixpoint.time=time ; iterations=iterations ; descendings=descendings ; stable=stable} =
    fprintf fmt "  time = %f@.  iterations = %d@.  descendings = %d@.  stable = %B@." time iterations descendings stable

  let print_stats_dot fmt {Fixpoint.time=time ; iterations=iterations ; descendings=descendings ; stable=stable} =
    fprintf fmt "time = %f\\niterations = %d" time iterations

  let print_abstract_state state_data_printer fmt v (s,av) ~pred ~succ =
    fprintf fmt "  { @[<hv 2>%a : @,%a@] }@." state_data_printer s (Lattice.print ()) av

  let print_abstract_output configuration state_data_printer transition_data_printer program_data_printer fmt output =
    let {LabeledTransitionSystem.graph = output_graph ; init_states = _init_states} = output in
    Graph.iter_vertex output_graph (print_abstract_state state_data_printer fmt) ;
    if Utils.Configuration.get_flag configuration "stats" then
      let p,stat = Graph.info output_graph in
        print_stats fmt stat

  let print_abstract_output_dot configuration state_data_printer transition_data_printer program_data_printer fmt output_graph =
    LabeledTransitionSystem.print_dot
      (fun fmt (s,av) -> fprintf fmt "%a" (Lattice.print ()) av)
      (fun fmt (t,g,c,av) -> match g with
        | True ->
            fprintf fmt "%a" print_commands c
        | _->
            fprintf fmt "{%a} %a" print_guard g print_commands c)
      (fun fmt (p,stat) -> ())
      fmt
      output_graph

  (* Creation of the output graph from those of MkFixpoint *)
  let create_abstract_graph
    (configuration : Utils.Configuration.t)
    (program_graph : ('s,'t,'p) input_graph)
    (lattice_manager : Lattice.manager)
    (output_graph : (abstract, arc, stat) Graph.t)
  =
    let f_vertex (v : vertex) s =
      let av = Graph.attrvertex output_graph v
      in (s, av)
    in
    let f_hedge (h : hedge) (t, g, c) =
      let av = Graph.attrhedge output_graph h
      in (t, g, c, av)
    in
    let f_info p =
      (p, Graph.info output_graph)
    in
      Graph.map program_graph f_vertex f_hedge f_info

  let run_and_print
    (configuration : Utils.Configuration.t)
    ({LabeledTransitionSystem.graph = program_graph ; init_states = init_states} : ('s,'t,'p) input)
    state_data_printer transition_data_printer program_data_printer fmt
  : unit
  =
    let lattice_manager = init_lattice_manager () in
    let output_graph = run_fixpoint_analysis configuration program_graph init_states lattice_manager in
    let abstract_graph =
    {
      LabeledTransitionSystem.
      graph = create_abstract_graph configuration program_graph lattice_manager output_graph ;
      init_states = init_states
    }
    in
    if Utils.Configuration.get_flag configuration "dot_output" then
      print_abstract_output_dot configuration state_data_printer transition_data_printer program_data_printer fmt abstract_graph
    else
      print_abstract_output configuration state_data_printer transition_data_printer program_data_printer fmt abstract_graph

end


module ZonesReachability = Reachability (Manage (Lattices.Dbm.Zdbm))
module MultisetReachability = Reachability (Manage (Lattices.Multiset.MultisetLattice))


