(* Bertrand Jeannet. This file is released under LGPL license. *)

(** Fixpoint analysis of an equation system: types *)

(** {2 Introduction}

    This module provides a generic engine for computing
    iteratively the solution of a fixpoint equation on a lattice.

    The equation system is represented with an hypergraph, in
    which vertices corresponds to unknown and oriented hyperedges
    to functions. It is assumed that hyperedges have unique
    destination vertex, and that associated functions are strict
    in each of their arguments: a bottom value in one of the
    argument implies that the result is empty.

    The equation system is parameterized by a manager, which
    provides the operations of the abstract lattices, and the
    transfer functions involved in the equation system.

    This interface is polymorphic and makes use of the following
    type variables:

    - ['vertex] is the type of vertex identifiers in the hypergraphs
    - ['hedge] is the type of hyperedges identifiers in the hypergraphs
    - ['abstract] is the type of abstract values (or dataflow properties)
    - ['arc] is the type of the information associated to hyperedges (optional features).

*)

open Format

(*  ********************************************************************** *)
(** {2 Public datatypes} *)
(*  ********************************************************************** *)

(*  ====================================================================== *)
(** {3 Manager} *)
(*  ====================================================================== *)

type ('vertex,'hedge,'abstract,'arc) manager = {
  (** Lattice operation *)
  mutable bottom : 'vertex -> 'abstract;
  mutable canonical : 'vertex -> 'abstract -> unit;
  mutable is_bottom : 'vertex -> 'abstract -> bool;
  mutable is_leq : 'vertex -> 'abstract -> 'abstract -> bool;
  mutable join :  'vertex -> 'abstract -> 'abstract -> 'abstract;
  mutable join_list : 'vertex -> 'abstract list -> 'abstract;
  mutable widening : 'vertex -> 'abstract -> 'abstract -> 'abstract;

  (** Initialisation of equations *)
  mutable abstract_init : 'vertex -> 'abstract;
  mutable arc_init : 'hedge -> 'arc;

  (** Interpreting hyperedges *)
  mutable apply : 'hedge -> 'abstract array -> 'arc * 'abstract;

  (** Printing functions *)
  mutable print_vertex : Format.formatter -> 'vertex -> unit;
  mutable print_hedge : Format.formatter -> 'hedge -> unit;
  mutable print_abstract: Format.formatter -> 'abstract -> unit;
  mutable print_arc: Format.formatter -> 'arc -> unit;

  (** Fixpoint Options *)
  mutable accumulate : bool;

  (** Widening Options *)
  mutable widening_first : bool;
  mutable widening_start : int;
  mutable widening_freq : int;
  mutable widening_descend : int;

  (** Printing Options *)
  mutable print_fmt : Format.formatter;
  mutable print_analysis : bool;
  mutable print_component : bool;
  mutable print_step : bool;
  mutable print_state : bool;
  mutable print_postpre : bool;
  mutable print_workingsets : bool;

  (** DOT Options *)
  mutable dot_fmt : Format.formatter option;
  mutable dot_vertex : Format.formatter -> 'vertex -> unit;
  mutable dot_hedge : Format.formatter -> 'hedge -> unit;
  mutable dot_attrvertex : Format.formatter -> 'vertex -> unit;
  mutable dot_attrhedge : Format.formatter -> 'hedge -> unit;
}

(*  ====================================================================== *)
(** {3 Dynamically explored equation system} *)
(*  ====================================================================== *)

type ('vertex,'hedge) equation =
  'vertex -> ('hedge, 'vertex array * 'vertex) Mappe.t

(*  ====================================================================== *)
(** {3 Iteration strategies} *)
(*  ====================================================================== *)

type ('vertex,'hedge) strategy_vertex = {
  mutable vertex : 'vertex;
  mutable hedges : 'hedge list;
  mutable widen : bool;
}
type ('vertex,'hedge) strategy = ('vertex,'hedge) strategy_vertex Ilist.t

let print_strategy_vertex
  (man:('vertex,'hedge,'abstract,'arc) manager)
  fmt
  (v:('vertex,'hedge) strategy_vertex)
  =
  fprintf fmt "@[<h>(%b,%a,%a)@]"
    v.widen
    man.print_vertex v.vertex
    (Print.list ~first:"@[<h>[" ~sep:"," ~last:"]@]"
	man.print_hedge)
    v.hedges

let print_strategy
  (man:('vertex,'hedge,'abstract,'arc) manager)
  fmt
  (s:('vertex,'hedge) strategy)
  =
  Ilist.print ~first:"[@[<hv>" (print_strategy_vertex man) fmt s


let make_strategy_vertex
    (compare:('vertex,'hedge) SHGraph.compare)
    (graph:('vertex,'hedge,'a,'b,'c) SHGraph.t)
    (widen:bool)
    (vertex:'vertex)
    :
    ('vertex,'hedge) strategy_vertex
    =
  let spredhedges = SHGraph.Compare.predhedge compare graph vertex in
  let hedges = Sette.elements spredhedges in
  let strategy_vertex = {
    vertex = vertex;
    hedges = hedges;
    widen = widen
  }
  in
  strategy_vertex

let make_strategy_default
    (compare:('vertex,'hedge) SHGraph.compare)
    ?(depth=10)
    ?priority
    ~(vertex_dummy:'vertex)
    ~(hedge_dummy:'hedge)
    (graph:('vertex,'hedge,'a,'b,'c) SHGraph.t)
    (sinit:'vertex Sette.t)
    :
    ('vertex,'hedge) strategy
    =
  let scfc =
    SHGraph.Compare.scfc_multi
      compare
      vertex_dummy hedge_dummy
      ?priority graph sinit
  in
  let scfc = Ilist.map
    (begin fun flag vertex ->
      make_strategy_vertex compare graph flag vertex
    end)
    scfc
  in
  Ilist.flatten ~depth scfc

(*  ====================================================================== *)
(** {3 Output} *)
(*  ====================================================================== *)

(** statistics at the end of the analysis *)
type stat = {
  time : float;
  iterations : int;
  descendings : int;
  stable : bool;
}

(** result of the analysis *)
type ('vertex,'hedge,'abstract,'arc) output =
  ('vertex,'hedge,'abstract, 'arc, stat) SHGraph.t

let print_stat fmt (stat:stat)
  =
  fprintf fmt "{ @[<hov>time=%f;@ iterations=%i;@ descendings=%i;@ stable=%b@] }"
    stat.time stat.iterations stat.descendings stat.stable

let print_output
    (compare:('vertex,'hedge) SHGraph.compare)
    (man:('vertex,'hedge,'abstract,'arc) manager)
    fmt
    (g:('vertex,'hedge,'abstract,'arc,stat) SHGraph.t)
    =
  SHGraph.Compare.print
    compare
    man.print_vertex
    man.print_hedge
    man.print_abstract
    man.print_arc
    print_stat
    fmt
    g

(*  ********************************************************************** *)
(** {2 Internal datatypes} *)
(*  ********************************************************************** *)

type 'abstract attr = {
  mutable reach : 'abstract;
  mutable first : 'abstract;
  mutable empty : bool;
}

type 'arc arc = {
  mutable arc : 'arc;
  mutable aempty : bool;
}

type ('vertex,'hedge) infodyn = {
  mutable iaddhedge : ('hedge,'vertex array * 'vertex) Hashhe.t;
  iequation : ('vertex,'hedge) equation
}

type ('vertex,'hedge) info = {
  iinit : 'vertex Sette.t;
  itime : float ref;
  iiterations : int ref;
  idescendings : int ref;
  istable : bool ref;
  mutable iworkvertex : ('vertex,unit) Hashhe.t;
  mutable iworkhedge : ('hedge,unit) Hashhe.t;
  iinfodyn : ('vertex,'hedge) infodyn option;
}

type ('vertex,'hedge,'abstract,'arc) graph =
  ('vertex, 'hedge, 'abstract attr, 'arc arc, ('vertex,'hedge) info) SHGraph.t

(*  ====================================================================== *)
(** {3 Printing functions} *)
(*  ====================================================================== *)

let print_attr
    (man:('vertex,'hedge,'abstract,'arc) manager)
    (fmt:Format.formatter)
    (attr:'abstract attr)
  =
  fprintf fmt "{ @[<hov>reach=%a;@ empty=%b@] }"
    man.print_abstract attr.reach
    attr.empty

let print_arc
    (man:('vertex,'hedge,'abstract,'arc) manager)
    (fmt:Format.formatter)
    (arc:'arc arc)
  =
  fprintf fmt "{ @[<hov>arc=%a;@ aempty=%b@] }"
    man.print_arc arc.arc
    arc.aempty

let print_info
    (man:('vertex,'hedge,'abstract,'arc) manager)
    (fmt:Format.formatter)
    (info:('vertex,'hedge) info)
  =
  fprintf fmt "{ @[<hov>itime=%f;@ iiterations=%i;@ idescendings=%i;@ istable=%b;@ iworkvertex=%a;@ iworkhedge=%a@] }"
    !(info.itime)
    !(info.iiterations)
    !(info.idescendings)
    !(info.istable)
    (Hashhe.print man.print_vertex (fun _ _ -> ())) info.iworkvertex
    (Hashhe.print man.print_hedge (fun _ _ -> ())) info.iworkhedge

let print_workingsets
    (man:('vertex,'hedge,'abstract,'arc) manager)
    (fmt:Format.formatter)
    (graph:('vertex,'hedge,'abstract,'arc) graph)
    :
    unit
    =
  let info = SHGraph.info graph in
  fprintf fmt "@[<v>workvertex = %a@ workhedge = %a"
    (Hashhe.print man.print_vertex (fun _ _ -> ())) info.iworkvertex
    (Hashhe.print man.print_hedge (fun _ _ -> ())) info.iworkhedge
  ;
  begin match info.iinfodyn with
  | Some(dyn) ->
      fprintf fmt "@ addhedge = %a"
	(Hashhe.print man.print_hedge (fun _ _ -> ()))
	dyn.iaddhedge
  | None -> ()
  end;
  fprintf fmt "@]"

let print_graph
    (compare:('vertex,'hedge) SHGraph.compare)
    (man:('vertex,'hedge,'abstract,'arc) manager)
    (fmt:Format.formatter)
    (g:('vertex,'hedge,'abstract,'arc) graph)
    =
  SHGraph.Compare.print
    compare
    man.print_vertex
    man.print_hedge
    (print_attr man)
    (print_arc man)
    (print_info man)
    fmt
    g

(*  ====================================================================== *)
(** {3 DOT functions} *)
(*  ====================================================================== *)

let set2_of_strategy
    (compare:('vertex,'hedge) SHGraph.compare)
    (strategy:('vertex,'hedge) strategy)
    :
    'vertex Sette.t * 'vertex Sette.t
    =
  Ilist.fold_left
    (begin fun (res,resw) _ v ->
      if v.widen then
	res,(Sette.Compare.add compare.SHGraph.comparev v.vertex resw)
      else
	(Sette.Compare.add compare.SHGraph.comparev v.vertex res),resw
    end)
    (Sette.empty,Sette.empty) strategy

let dot_graph
    (compare:('vertex,'hedge) SHGraph.compare)
    ?(style="size=\"7.5,10\";center=true;ranksep=0.16;nodesep=0.2;")
    ?titlestyle
    ?(vertexstyle="shape=box,fontsize=10,height=0.2,width=0.1")
    ?(hedgestyle="shape=plaintext,fontsize=12,height=0.0,width=0.1")
    ?strategy
    ?vertex
    (man:('vertex,'hedge,'abstract,'arc) manager)
    (graph:('vertex,'hedge,'abstract,'arc) graph)
    ~(title:string)
    =
  match man.dot_fmt with
  | None -> ()
  | Some dot_fmt ->
      let (set,setw) = match strategy with
	| None -> Sette.empty,Sette.empty
	| Some strategy -> set2_of_strategy compare strategy
      in
      let info = SHGraph.info graph in
      let fvertexstyle = begin fun v ->
	let vertexstyle =
	  if begin match vertex with
	  | None -> false
	  | Some vertex -> (compare.SHGraph.comparev v vertex) = 0
	  end
	  then
	    vertexstyle^",style=filled,fillcolor=coral1"
	  else if Sette.Compare.mem compare.SHGraph.comparev v setw then
	    vertexstyle^",style=filled,fillcolor=red1"
	  else if Sette.Compare.mem compare.SHGraph.comparev v set then
	    vertexstyle^",style=filled,fillcolor=orange1"
	  else
	    vertexstyle
	in
	let vertexstyle =
	  if Hashhe.Compare.mem compare.SHGraph.hashv info.iworkvertex v then
	    vertexstyle^",fontcolor=blue3"
	  else
	    vertexstyle
	in
	vertexstyle
      end
      in
      let fhedgestyle = begin fun h ->
	let hedgestyle =
	  if Hashhe.Compare.mem compare.SHGraph.hashh info.iworkhedge h then
	    hedgestyle^",fontcolor=blue3"
	  else
	    hedgestyle
	in
	hedgestyle
      end
      in
      SHGraph.print_dot
	~style
	~fvertexstyle
	~fhedgestyle
	~title
	(fun fmt vertex ->
	  fprintf fmt "\"%s\""
	    (Print.escaped ~linebreak:'n'
	      (Print.sprintf "%a" man.dot_vertex vertex)))
	(fun fmt hedge ->
	  fprintf fmt "\"%s\""
	    (Print.escaped ~linebreak:'n'
	      (Print.sprintf "%a" man.dot_hedge hedge)))
	(fun fmt vertex attr ->
	  fprintf fmt "%s"
	    (Print.escaped ~linebreak:'n'
	      (Print.sprintf ~margin:40 "%a@.%a"
		man.dot_attrvertex vertex
		man.print_abstract attr.reach)))
	(fun fmt hedge arc ->
	  fprintf fmt "%s"
	    (Print.escaped ~linebreak:'n'
	      (Print.sprintf ~margin:40 "%a@.%a"
		man.dot_attrhedge hedge
		man.print_arc arc.arc)))
	dot_fmt
	graph
