
open Types
open Domain
open Identificator

(* TODO :
 * - StdExpression can be viewed as SymbolicExpressions using StdIdentificators
 * - Expressions should be well typed trees
 *     -> Dissociate Plus/FPlus, Mul/FMul, for integer and floating point in expression
 *        and fexpression and add ToFloat and ToInt constructor
 *     -> Value has pretty no sense, should be lvalue, containing variable, array access
 *        or dereferenciation
 *     -> Boolean predicates may then be lvalues
 *     -> Commands may reflect types, and there could be IntegerAssign, BooleanAssign
 *        FloatAssign, Adress assign, etc.
 * - Lack of Mod and some binary operations to be able to fully transform C AST to
 *   flow graph.
 * - Some more iterators may be useful in general
 *)

(**
 *  Constants used in expressions
 *)

type const = Integer of int | Rational of Rational.t
module Constants : Domain  with type t = const


(**
 *  Expressions types
 *
 *  Cover all the expressions usable in programs
 *)

module StdExpression :
sig
  (* Polymorphic types of expression *)

  type 's p_value = 
    | Const of const
    | Array of 's * 's p_expression
    | Variable of 's
    | Unknown (* Non deterministic value *)
  and 's p_expression =
    | Value of 's p_value
    | Neg of 's p_expression
    | Plus of 's p_expression * 's p_expression
    | Minus of 's p_expression * 's p_expression
    | Mul of 's p_expression * 's p_expression
    | Div of 's p_expression * 's p_expression

  type 's p_comparison =
    | Eq of 's p_expression * 's p_expression
    | Neq of 's p_expression * 's p_expression
    | GreaterEq of 's p_expression * 's p_expression
    | Greater of 's p_expression * 's p_expression

  type 's p_formula =
    | Not of 's p_formula
    | Or of 's p_formula * 's p_formula
    | And of 's p_formula * 's p_formula
    | Imply of 's p_formula * 's p_formula
    | True | False
    | Predicate of 's (* Boolean variable *)
    | Comparison of 's p_comparison

  (* Commands and guards *)

  type 'expression generic_guard = 'expression
  type ('value,'formula,'expression) generic_command =
    | BooleanAssign of 'value * 'formula
    | NumericalAssign of 'value * 'expression

  type 's p_guard = ('s p_formula) generic_guard
  type 's p_command = ('s p_value,'s p_formula,'s p_expression) generic_command

  (* Iterators *)

  val value_map : ('a -> 'b) -> 'a p_value -> 'b p_value
  val expression_map : ('a -> 'b) -> 'a p_expression -> 'b p_expression
  val comparison_map : ('a -> 'b) -> 'a p_comparison -> 'b p_comparison
  val formula_map : ('a -> 'b) -> 'a p_formula -> 'b p_formula
  val guard_map : ('a -> 'b) -> 'a p_guard -> 'b p_guard
  val command_map : ('a -> 'b) -> 'a p_command -> 'b p_command

  val expression_map_values : ('a p_value -> 'b p_value) -> 'a p_expression -> 'b p_expression
  val comparison_map_values : ('a p_value -> 'b p_value) -> 'a p_comparison -> 'b p_comparison
  val formula_map_values : ('a p_value -> 'b p_value) -> 'a p_formula -> 'b p_formula
  val guard_map_values : ('a p_value -> 'b p_value) -> 'a p_guard -> 'b p_guard
  val command_map_values : ('a p_value -> 'b p_value) -> 'a p_command -> 'b p_command


  (* Standard expressions using symbol *)

  type value = symbol p_value
  type expression = symbol p_expression
  type comparison = symbol p_comparison
  type formula = symbol p_formula
  type guard = symbol p_guard
  type command = symbol p_command

  val print_value : value printer
  val print_expression : expression printer
  val print_comparison : comparison printer
  val print_formula : formula printer
  val print_guard : guard printer
  val print_command : command printer
  val print_commands : (command list) printer
end

open StdExpression


(*
 * Symbolic expressions based on arbitrary identificators
 *)

module SymbolicExpression (I : Identificator) :
sig
  type symbol = I.t

  type value = symbol p_value
  type expression = symbol p_expression
  type comparison = symbol p_comparison
  type formula = symbol p_formula
  type guard = symbol p_guard
  type command = symbol p_command

  (* Comparison function *)
  val compare_value : value -> value -> int
  val compare_expression : expression -> expression -> int
  val compare_comparison : comparison -> comparison -> int
  val compare_formula : formula -> formula -> int

  (* Expression conversion *)
  val from_value : StdExpression.value -> value
  val from_expression : StdExpression.expression -> expression
  val from_comparison : StdExpression.comparison -> comparison
  val from_formula : StdExpression.formula -> formula

  (* Printing *)
  val print_value : value printer
  val print_expression : expression printer
  val print_comparison : comparison printer
  val print_formula : formula printer

  module Guard :
  sig
    type t = guard
    val from_guard : StdExpression.guard -> t
  end

  module Command :
  sig
    type t = command
    val from_command : StdExpression.command -> t
  end
end


(**
 *  Canonical expressions
 *
 *  used to simplify semantics work by reducing the number of cases to consider
 *)

(* Linear expression allow to forget associativity and commutativity in sum or products
   and use monadic comparison operators 

   In linear expression, there is no negation nor multiplication by a constant. All theses
   are converted to Sum. This invariant is used to perform domain operation. Linear Expressions
   are not a domain yet, but will be one sunny day.
*)

module LinearExpression (I : Identificator) :
sig
  type symbol = I.t

  type value =
    | Const of const
    | Array of symbol * expression
    | Variable of symbol
    | Unknown
  and expression =
    | Value of value
    | Sum of (const * expression) list * const
    | Mul of expression * expression
    | Div of expression * expression

  type comparison =
    | Nul of expression
    | NotNul of expression
    | PositiveNul of expression
    | Positive of expression

  type formula =
    | Not of formula 
    | Or of formula * formula
    | And of formula * formula
    | True | False
    | Predicate of symbol
    | Comparison of comparison

  type guard = formula generic_guard
  type command = (value,formula,expression) generic_command


  (* Comparison function for ordering these expressions *)
  val compare_value : value -> value -> int
  val compare_expression : expression -> expression -> int
  val compare_comparison : comparison -> comparison -> int
  val compare_formula : formula -> formula -> int

  (* Printing *)
  val print_value : value printer
  val print_expression : expression printer
  val print_comparison : comparison printer
  val print_formula : formula printer

  (* Expression conversion *)
  val from_symbolic_value : symbol p_value -> value
  val from_symbolic_expression : symbol p_expression -> expression
  val from_symbolic_comparison : symbol p_comparison -> comparison
  val from_symbolic_formula : symbol p_formula -> formula

  val from_value : StdExpression.value -> value
  val from_expression : StdExpression.expression -> expression
  val from_comparison : StdExpression.comparison -> comparison
  val from_formula : StdExpression.formula -> formula

  (* Domain operations *)
  val invers : expression -> expression
  val plus : expression -> expression -> expression
  val minus : expression -> expression -> expression

  (* Other operations *)
  val get_symbols_from_numerical_expression : expression -> symbol list
  val substitute_term : expression -> expression -> expression -> expression
  val solve : expression -> value -> expression

  module Guard :
  sig
    type t = guard
    val from_guard : StdExpression.guard -> t
    val from_symbolic_guard : symbol p_guard -> t
  end

  module Command :
  sig
    type t = command
    val from_command : StdExpression.command -> t
    val from_symbolic_command : symbol p_command -> t
  end
end

