/**
  Syntax Analysis for C files

  from http://www.lysator.liu.se/c/ANSI-C-grammar-y.html
*/

%{
  open CAbstractSyntaxTree
%}

%token EOF

%token <Types.symbol> IDENTIFIER
%token <Expr.const> CONSTANT
%token <Types.symbol> TYPE_NAME
%token <string> STRING_LITERAL

%token SIZEOF
%token PTR_OP INC_OP DEC_OP LEFT_OP RIGHT_OP LE_OP GE_OP EQ_OP NE_OP
%token AND_OP OR_OP MUL_ASSIGN DIV_ASSIGN MOD_ASSIGN ADD_ASSIGN
%token SUB_ASSIGN LEFT_ASSIGN RIGHT_ASSIGN AND_ASSIGN
%token XOR_ASSIGN OR_ASSIGN

%token SEMICOLON LBRACE RBRACE COMMA COLON  EQUAL LPAR RPAR
%token LSQUARE RSQUARE DOT AMP EXCLAMATION TILDE MINUS PLUS STAR
%token SLASH PERCENT LT_OP GT_OP HAT PIPE QUESTION

%token TYPEDEF EXTERN STATIC AUTO REGISTER
%token CHAR SHORT INT LONG SIGNED UNSIGNED FLOAT DOUBLE CONST VOLATILE VOID
%token STRUCT UNION ENUM ELLIPSIS

%token ASSUME CASE DEFAULT IF ELSE SWITCH WHILE DO FOR GOTO CONTINUE BREAK RETURN

%type <CAbstractSyntaxTree.program> file
%start file

%%

file
        : translation_unit EOF { $1 }
        | EOF { [] }
        ;

primary_expression
	: IDENTIFIER { Variable $1 }
	| CONSTANT { Constant $1 }
	| STRING_LITERAL { String $1 }
	| LPAR expression RPAR { $2 }
	;

postfix_expression
	: primary_expression { $1 }
	| postfix_expression LSQUARE expression RSQUARE { ArrayAccess($1, $3) }
	| postfix_expression LPAR RPAR { FunctionCall($1, []) }
	| postfix_expression LPAR argument_expression_list RPAR { FunctionCall($1, $3) }
	| postfix_expression DOT IDENTIFIER { Selection($1, $3) }
	| postfix_expression PTR_OP IDENTIFIER { PointerSelection($1, $3) }
	| postfix_expression INC_OP { PostIncrementation($1) }
	| postfix_expression DEC_OP { PostDecrementation($1) }
	;

argument_expression_list
	: assignment_expression { [$1] }
	| argument_expression_list COMMA assignment_expression { $1 @ [$3] }
	;

unary_expression
	: postfix_expression { $1 }
	| INC_OP unary_expression { PreIncrementation $2 }
	| DEC_OP unary_expression { PreDecrementation $2 }
	| AMP cast_expression { Adress $2 }
	| STAR cast_expression { Dereferenciation $2 }
	| PLUS cast_expression { $2 }
	| MINUS cast_expression { Neg $2 }
	| TILDE cast_expression { BinaryNot $2 }
	| EXCLAMATION cast_expression { LogicalNot $2 }
	| SIZEOF unary_expression { SizeofExpression $2 }
	| SIZEOF LPAR type_name RPAR { SizeofType $3 }
	;

cast_expression
	: unary_expression { $1 }
	| LPAR type_name RPAR cast_expression { Cast ($2, $4) }
	;

multiplicative_expression
	: cast_expression { $1 }
	| multiplicative_expression STAR cast_expression { Mul ($1, $3) }
	| multiplicative_expression SLASH cast_expression { Div ($1, $3) }
	| multiplicative_expression PERCENT cast_expression { Mod ($1, $3) }
	;

additive_expression
	: multiplicative_expression { $1 }
	| additive_expression PLUS multiplicative_expression { Add ($1, $3) }
	| additive_expression MINUS multiplicative_expression { Sub ($1, $3) }
	;

shift_expression
	: additive_expression { $1 }
	| shift_expression LEFT_OP additive_expression { ShiftLeft ($1, $3) }
	| shift_expression RIGHT_OP additive_expression { ShiftRight ($1, $3) }
	;

relational_expression
	: shift_expression { $1 }
	| relational_expression LT_OP shift_expression { Lower ($1, $3) }
	| relational_expression GT_OP shift_expression { Greater ($1, $3) }
	| relational_expression LE_OP shift_expression { LowerEq ($1, $3) }
	| relational_expression GE_OP shift_expression { GreaterEq ($1, $3) }
	;

equality_expression
	: relational_expression { $1 }
	| equality_expression EQ_OP relational_expression { Eq ($1, $3) }
	| equality_expression NE_OP relational_expression { Neq ($1, $3) }
	;

and_expression
	: equality_expression { $1 }
	| and_expression AMP equality_expression { BinaryAnd ($1, $3) }
	;

exclusive_or_expression
	: and_expression { $1 }
	| exclusive_or_expression HAT and_expression { BinaryXor ($1, $3) }
	;

inclusive_or_expression
	: exclusive_or_expression { $1 }
	| inclusive_or_expression PIPE exclusive_or_expression { BinaryOr ($1, $3) }
	;

logical_and_expression
	: inclusive_or_expression { $1 }
	| logical_and_expression AND_OP inclusive_or_expression { LogicalAnd ($1, $3) }
	;

logical_or_expression
	: logical_and_expression { $1 }
	| logical_or_expression OR_OP logical_and_expression { LogicalOr ($1, $3) }
	;

conditional_expression
	: logical_or_expression { $1 }
	| logical_or_expression QUESTION expression COLON conditional_expression { Conditional ($1, $3, $5) }
	;

assignment_expression
	: conditional_expression { $1 }
	| unary_expression EQUAL assignment_expression { Assign ($1, $3) }
	| unary_expression MUL_ASSIGN assignment_expression { MulAssign ($1, $3) }
	| unary_expression DIV_ASSIGN assignment_expression { DivAssign ($1, $3) }
	| unary_expression MOD_ASSIGN assignment_expression { ModAssign ($1, $3) }
	| unary_expression ADD_ASSIGN assignment_expression { AddAssign ($1, $3) }
	| unary_expression SUB_ASSIGN assignment_expression { SubAssign ($1, $3) }
	| unary_expression LEFT_ASSIGN assignment_expression { LeftShiftAssign ($1, $3) }
	| unary_expression RIGHT_ASSIGN assignment_expression { RightShiftAssign ($1, $3) }
	| unary_expression AND_ASSIGN assignment_expression { AndAssign ($1, $3) }
	| unary_expression XOR_ASSIGN assignment_expression { XorAssign ($1, $3) }
	| unary_expression OR_ASSIGN assignment_expression { OrAssign ($1, $3) }
	;

expression
	: assignment_expression { $1 }
	| expression COMMA assignment_expression { Sequence($1, $3) }
	;

constant_expression
	: conditional_expression { $1 }
	;

declaration
	: declaration_specifiers SEMICOLON { $1, [] }
	| declaration_specifiers init_declarator_list SEMICOLON { $1, $2 }
	;

declaration_specifiers
	: storage_class_specifier { [$1], [], [] }
	| storage_class_specifier declaration_specifiers { let lsc, ls, lq = $2 in $1 :: lsc, ls, lq }
	| type_specifier { [], [$1], [] }
	| type_specifier declaration_specifiers { let lsc, ls, lq = $2 in lsc, $1 :: ls, lq }
	| type_qualifier { [], [], [$1] }
	| type_qualifier declaration_specifiers { let lsc, ls, lq = $2 in lsc, ls, $1 :: lq }
	;

init_declarator_list
	: init_declarator { [$1] }
	| init_declarator_list COMMA init_declarator { $1 @ [$3] }
	;

init_declarator
	: declarator { let s,d =$1 in s,d,None }
	| declarator EQUAL init { let s,d = $1 in s,d,Some $3 }
	;

storage_class_specifier
	: TYPEDEF { Typedef }
	| EXTERN { Extern }
	| STATIC { Static }
	| AUTO { Auto }
	| REGISTER { Register }
	;

type_specifier
	: VOID { Void}
	| CHAR { Char }
	| SHORT { Short }
	| INT { Int }
	| LONG { Long }
	| FLOAT { Float }
	| DOUBLE { Double }
	| SIGNED { Signed }
	| UNSIGNED { Unsigned }
	| struct_or_union_specifier { $1 }
	| enum_specifier { $1 }
	| TYPE_NAME { UserType $1 }
	;

struct_or_union_specifier
	: STRUCT IDENTIFIER LBRACE struct_declaration_list RBRACE { Struct (Some $2, Some $4) }
	| STRUCT LBRACE struct_declaration_list RBRACE { Struct (None, Some $3) }
	| STRUCT IDENTIFIER { Struct (Some $2, None) }
	| UNION IDENTIFIER LBRACE struct_declaration_list RBRACE { Union (Some $2, Some $4) }
	| UNION LBRACE struct_declaration_list RBRACE { Union (None, Some $3) }
	| UNION IDENTIFIER { Struct (Some $2, None) }
	;

struct_declaration_list
	: struct_declaration { [$1] }
	| struct_declaration_list struct_declaration { $1 @ [$2] }
	;

struct_declaration
	: specifier_qualifier_list struct_declarator_list SEMICOLON { let t,q = $1 in t,q,$2 }
	;

specifier_qualifier_list
	: type_specifier specifier_qualifier_list { let ls, lq = $2 in $1 :: ls, lq }
	| type_specifier { [$1], [] }
	| type_qualifier specifier_qualifier_list { let ls, lq = $2 in ls, $1 :: lq }
	| type_qualifier { [], [$1] }
	;

struct_declarator_list
	: struct_declarator { [$1] }
	| struct_declarator_list COMMA struct_declarator { $1 @ [$3] }
	;

struct_declarator
	: declarator { let s,d=$1 in Some s, d, None }
	| COLON constant_expression { None, Scalar, Some $2 }
	| declarator COLON constant_expression { let s,d=$1 in Some s, d, Some $3 }
	;

enum_specifier
	: ENUM LBRACE enumerator_list RBRACE { Enum (None, Some $3) }
	| ENUM IDENTIFIER LBRACE enumerator_list RBRACE { Enum (Some $2, Some $4) }
	| ENUM IDENTIFIER { Enum (Some $2, None) }
	;

enumerator_list
	: enumerator { [$1] }
	| enumerator_list COMMA enumerator { $1 @ [$3] }
	;

enumerator
	: IDENTIFIER { $1, None }
	| IDENTIFIER EQUAL constant_expression { $1, Some $3 }
	;

type_qualifier
	: CONST { Const }
	| VOLATILE { Volatile }
	;

declarator
	: direct_declarator { $1 }
	| STAR declarator { let s,d = $2 in s, Pointer ([], d) }
	| STAR type_qualifier_list declarator { let s,d = $3 in s, Pointer ($2, d) }
	;

direct_declarator
	: IDENTIFIER { $1, Scalar }
	| LPAR declarator RPAR { $2 }
	| direct_declarator LSQUARE constant_expression RSQUARE { let s,d=$1 in s, Array (d, Some $3) }
	| direct_declarator LSQUARE RSQUARE { let s,d=$1 in s,Array(d, None) } 
	| direct_declarator LPAR parameter_type_list RPAR { let s,d=$1 in s, Function (d, $3) }
/*	| direct_declarator LPAR identifier_list RPAR */
	| direct_declarator LPAR RPAR { let s,d =$1 in s, Function (d, []) }
	;

type_qualifier_list
	: type_qualifier { [$1] }
	| type_qualifier_list type_qualifier { $1 @ [$2] }
	;


parameter_type_list
	: parameter_list { $1 }
	| parameter_list COMMA ELLIPSIS { $1 @ [Ellipsis] }
	;

parameter_list
	: parameter_declaration { [Parameter $1] }
	| parameter_list COMMA parameter_declaration { $1 @ [Parameter $3] }
	;

parameter_declaration
	: declaration_specifiers declarator { let s,d = $2 in Some s, $1, d }
	| declaration_specifiers abstract_declarator { None, $1, $2 }
	| declaration_specifiers { None, $1, Scalar }
	;
/*
identifier_list
	: IDENTIFIER { [$1] }
	| identifier_list COMMA IDENTIFIER { $1 @ [$3] }
	;
*/
type_name
	: specifier_qualifier_list { let s,q = $1 in s,q,Scalar }
	| specifier_qualifier_list abstract_declarator { let s,q = $1 in s,q,$2 }
	;

abstract_declarator
	: direct_abstract_declarator { $1 }
	| STAR { Pointer ([], Scalar) }
	| STAR type_qualifier_list { Pointer ($2, Scalar) }
	| STAR abstract_declarator { Pointer ([], $2) }
	| STAR type_qualifier_list abstract_declarator { Pointer ($2, $3) }
	;

direct_abstract_declarator
	: LPAR abstract_declarator RPAR { $2 }
	| LSQUARE RSQUARE { Array(Scalar, None) }
	| LSQUARE constant_expression RSQUARE { Array(Scalar, Some $2) }
	| direct_abstract_declarator LSQUARE RSQUARE { Array($1, None) }
	| direct_abstract_declarator LSQUARE constant_expression RSQUARE { Array($1, Some $3) }
	| LPAR RPAR { Function(Scalar, []) }
	| LPAR parameter_type_list RPAR { Function(Scalar, $2) }
	| direct_abstract_declarator LPAR RPAR { Function($1, []) }
	| direct_abstract_declarator LPAR parameter_type_list RPAR { Function($1, $3) }
	;

init
	: assignment_expression { ScalarValue $1 }
	| LBRACE init_list RBRACE { ArrayValue $2 }
	| LBRACE init_list COMMA RBRACE { ArrayValue $2 }
	;

init_list
	: init { [$1] }
	| init_list COMMA init { $1 @ [$3] }
	;

statement
	: assume_statement { $1 }
        | labeled_statement { $1 }
	| compound_statement { $1 }
	| expression_statement { match $1 with None -> Nop | Some e -> Expression e }
	| selection_statement { $1 }
	| iteration_statement { $1 }
	| jump_statement { $1 }
	;

assume_statement
	: ASSUME LPAR expression RPAR SEMICOLON { Assume $3 }

labeled_statement
	: IDENTIFIER COLON statement { LabelledStatement (Label $1, $3) }
	| CASE constant_expression COLON statement { LabelledStatement (Case $2, $4) }
	| DEFAULT COLON statement { LabelledStatement (Default, $3) }
	;

compound_statement
	: LBRACE RBRACE { Compound([], []) }
	| LBRACE statement_list RBRACE { Compound([], $2) }
	| LBRACE declaration_list RBRACE { Compound($2, []) }
	| LBRACE declaration_list statement_list RBRACE { Compound($2, $3) }
	;

declaration_list
	: declaration { [$1] }
	| declaration_list declaration { $1 @ [$2] }
	;

statement_list
	: statement { [$1] }
	| statement_list statement { $1 @ [$2] }
	;

expression_statement
	: SEMICOLON { None }
	| expression SEMICOLON { Some $1 }
	;

selection_statement
	: IF LPAR expression RPAR statement { If ($3, $5, Nop) }
	| IF LPAR expression RPAR statement ELSE statement { If ($3, $5, $7) }
	| SWITCH LPAR expression RPAR statement { Switch ($3, $5) }
	;

iteration_statement
	: WHILE LPAR expression RPAR statement { While ($3, $5) }
	| DO statement WHILE LPAR expression RPAR SEMICOLON { Do ($5, $2) }
	| FOR LPAR expression_statement expression_statement RPAR statement { For ($3, $4, None, $6) }
	| FOR LPAR expression_statement expression_statement expression RPAR statement { For ($3, $4, Some $5, $7) }
	;

jump_statement
	: GOTO IDENTIFIER SEMICOLON { Goto($2) }
	| CONTINUE SEMICOLON { Continue }
	| BREAK SEMICOLON { Break }
	| RETURN SEMICOLON { Return (None) }
	| RETURN expression SEMICOLON { Return (Some ($2)) }
	;

translation_unit
	: external_declaration { [$1] }
	| translation_unit external_declaration { $1 @ [$2] }
	;

external_declaration
	: function_definition { FunctionDefinition $1 }
	| declaration { Declaration $1 }
	;

function_definition
/*	: declaration_specifiers declarator declaration_list compound_statement */
	: declaration_specifiers declarator compound_statement { let s,d = $2 in s,$1,d,$3 }
/*	| declarator declaration_list compound_statement */
	| declarator compound_statement { let s,d = $1 in s,([],[],[]),d,$2 }
	;
