
open Format

open Utils
open Types
open Expr.StdExpression
open FastModel


module Compiler : Compiler.Compiler =
struct
  module ParsingFunction =
  struct
    type token = FastParser.token
    type ast = program
    let the_lexer = FastLexer.lexBac
    let the_parser = FastParser.prog
  end

  module Parser = Parser.MakeParser (ParsingFunction)
  module Graph = LabeledTransitionSystem.Graph


  (* Data stored into the graph - Could carry line numbers, and other things *)

  type state_data = string
  type transition_data = string
  type program_data = ProgramFile of string | Nothing


  (* The graph to produce *)

  type program_graph = (state_data, transition_data*guard*(command list), program_data) LabeledTransitionSystem.t


  (* Programs manipulation *)

  let regions (prog : program) : (symbol * region) list =
    let filter acc = function
      | Declaration (Region (s,r)) -> (s,r)::acc
      | _ -> acc
    in List.fold_left filter [] prog.strategy.instlist

  let rearrange_prog (prog : program) : unit =
    (* the model is mutable, we reverse lists *)
    prog.model.vars <- List.rev prog.model.vars;
    prog.model.arrays <- List.rev prog.model.arrays;
    prog.model.states <- List.rev prog.model.states;
    (* all the actions have already been reversed during the parsing *)
    prog.model.transitions <- List.rev prog.model.transitions


  (* Transformation program -> graph *)

  (* add a vertex from symbol *)
  let add_vertex graph (s : symbol) =
    let state_data = Symbol.to_string s in
    Graph.add_vertex graph s state_data

  (* add an hyperedge from transition t*)
  let add_hedge graph (t : transition) =
    try (
      let transition_data = Symbol.to_string t.ident in
      Graph.add_hedge graph t.ident (transition_data, t.guard, t.actions) ~pred:[|t.fromid|] ~succ:[|t.toid|]
    )
    with Failure _ ->
      let s = sprintf "states '%s' or '%s' not declared" (Symbol.to_string t.fromid) (Symbol.to_string t.toid) in 
      raise (SemanticError s)

  let create_graph (prog : program) (data : program_data) : program_graph =
    rearrange_prog prog ;
    let states = prog.model.states in
    let transitions = prog.model.transitions in
    let states_count = List.length states in
    let transitions_count = List.length transitions in
    let hashtables_size = 2 * (states_count + transitions_count) in
    let graph = Graph.create hashtables_size data in
    List.iter (add_vertex graph) states ;
    List.iter (add_hedge graph) transitions ;
    (* TODO : a real init set *)
    let init_state = Symbol.add "init" in
    if Graph.is_vertex graph init_state then
    {
      LabeledTransitionSystem.graph = graph ;
      LabeledTransitionSystem.init_states = Graph.SetV.singleton init_state
    }
    else
      raise (SemanticError "no init state")

  let compile_file (program_file : string) : program_graph =
    let prog = Parser.parse_file program_file in
    create_graph prog (ProgramFile program_file)

  let compile_string (program_string : string) : program_graph =
    let prog = Parser.parse_string program_string in
    create_graph prog Nothing

  let print_state_data fmt state_data =
    fprintf fmt "%s" state_data

  let print_transition_data fmt transition_data =
    fprintf fmt "%s" transition_data

  let print_program_data fmt = function
    | ProgramFile file_name -> fprintf fmt "%s" (Filename.chop_extension (Filename.basename file_name))
    | Nothing -> fprintf fmt ""

  let print_dot (fmt : formatter) (graph : program_graph) : unit =
    LabeledTransitionSystem.print_dot
      print_state_data
      (fun fmt (t,g,c) -> fprintf fmt "%a {%a} %a" print_transition_data t print_guard g print_commands c)
      print_program_data
      fmt
      graph
end
