
open Types
open Expr.StdExpression
open Identificator


type semantic_result = Exact | Abstracted | Skip

(* Combine two semantic result used to compute an abstract value *)
val combine_semantic_result : semantic_result -> semantic_result -> semantic_result



module type SemiLattice =
sig
  type t
    (** Type of abstract values *)

  (* Complete lattices constants *)

  val bottom : unit -> t
    (** Create a bottom value *)
  val top : unit -> t
    (** Create a top value *)
  val is_bottom : t -> bool
    (** Return true if abstract value is bottom *)


  (* Lattice order and operators *)

  val is_leq : t -> t -> bool
    (** Inclusion test *)
  val join : t -> t -> t
  val join_list : t list -> t
    (** Binary and n-ary join operation *)

  val canonical : t -> unit
    (** Make an abstract value canonical *)

  val widening : t -> t -> t
    (** Classical windening operator *)

  (* Abstract domain *)
  module Semantics :
  sig
    module Guard :
    sig
      type t
      val from_guard : guard -> t
    end

    module Command :
    sig
      type t
      val from_command : command -> t
    end

    val apply_guard : t -> Guard.t -> t * semantic_result
      (** Return an abstract value complying with the guard *)
    val apply_command : t -> Command.t -> t * semantic_result
      (** Return the abstract value where the command has been applied  *)
  end

  (* Printing *)

  val print : t printer
    (** Print an abstract value *)
end


module type Lattice =
sig
  include SemiLattice

  val meet : t -> t -> t
  val meet_list : t list -> t
    (** Binary and n-ary meet operation *)
end


module type ManagedSemiLattice =
sig
  type t
    (** Type of abstract values *)
  type manager
    (** Type of manager *)

  (* Complete lattices constants *)

  val bottom : manager -> t
    (** Create a bottom value *)
  val top : manager -> t
    (** Create a top value *)
  val is_bottom : manager -> t -> bool
    (** Return true if abstract value is bottom *)


  (* Lattice order and operators *)

  val is_leq : manager -> t -> t -> bool
    (** Inclusion test *)
  val join : manager -> t -> t -> t
  val join_list : manager -> t list -> t
    (** Binary and n-ary join operation *)

  val canonical : manager -> t -> unit
    (** Make an abstract value canonical *)

  val widening : manager -> t -> t -> t
    (** Classical windening operator *)

  (* Abstract domain *)
  module Semantics :
  sig
    module Guard :
    sig
      type t
      val from_guard : manager -> guard -> t
    end

    module Command :
    sig
      type t
      val from_command : manager -> command -> t
    end

    val apply_guard : manager -> t -> Guard.t -> t * semantic_result
      (** Return an abstract value complying with the guard *)
    val apply_command : manager -> t -> Command.t -> t * semantic_result
      (** Return the abstract value where the command has been applied  *)
  end


  (* Printing *)

  val print : manager -> t printer
    (** Print an abstract value *)
end

module type ManagedLattice =
sig
  include ManagedSemiLattice

  val meet : manager -> t -> t -> t
  val meet_list : manager -> t list -> t
    (** Binary and n-ary meet operation *)
end

module type DimensionalLattice =
sig
  include Lattice

  module Var : Identificator
    (** Type of identificators *)

  val dimensions : t -> Var.t list
  val dimension_add : t -> Var.t -> t
  val dimension_remove : t -> Var.t -> t
  val dimension_rename : t -> (Var.t * Var.t) -> t
end



module Manage (L : Lattice) : ManagedLattice
  with type manager = unit
   and type t = L.t
   and type Semantics.Guard.t = L.Semantics.Guard.t
   and type Semantics.Command.t = L.Semantics.Command.t



(**
 * Add bottom to a non-complete lattice
 *
 * Since Bottom exists only in the extended Lattice :
 * - The original lattice don't actually have to implement bottom ()
 * - The original lattice can't force the value to be canonize to bottom. This
 *   is done when canonical is called through the is_bottom function from the
 *   original lattice.
 *)

type 't completed_abstract_value = Bottom | Abs of 't

module AddBottom (L : Lattice) : Lattice
  with type t = (L.t completed_abstract_value) ref
   and module Semantics.Guard = L.Semantics.Guard
   and module Semantics.Command = L.Semantics.Command

