
open Format
open Array

open Types
open Identificator
open Domain
open Lattice
open SquareMatrix


(**
 *  Difference Bound Matrices Abstract Domain
*)

module AddSpecialIdentificators (Base: Identificator) =
struct
  module OrderedId =
  struct
    type t = Normal of Base.t | Zero | Command

    let compare i1 i2 = match i1, i2 with
    | Zero, Zero -> 0
    | Zero, _ -> -1
    | _, Zero -> 1
    | Command, Command -> 0
    | Command, _ -> -1
    | _, Command -> 1
    | Normal id1, Normal id2 -> Base.compare id1 id2
  end

  include OrderedId

  module SetteId = Sette.Make(OrderedId)
  module MappeId = Mappe.Make(SetteId)

  let print fmt = function
    | Zero -> fprintf fmt "0"
    | Command -> fprintf fmt "1"
    | Normal id -> Base.print fmt id

  let from_base s = Normal s
  let from_symbol s = Normal (Base.from_symbol s)
end


(** DBM lattice over Integers *)

module ZDbmLattice (BaseIdentificator: Identificator) : DimensionalLattice =
struct
  (** Types *)

  module Space = AddInfinitiesToGroup (Integers)
  module Var = AddSpecialIdentificators (BaseIdentificator)
  module Mat = MappeSquareMatrix (Var)
  (* module Mat = StandardSquareMatrix (Var) *) (* Incompatible since it does not ensure variables are in the same order *)
  module Graph = Mat.Graph (Space)

  open Space
  open Var

  (* a DBM is a matrix of space element *)
  type t_dbm = Abs of (Space.t Mat.t) | Bottom
  type t = t_dbm ref


  (* Printing *)

  let print_internal fmt av =
    fprintf fmt "\n------------------------------------------\n";
    begin
      match !av with
        | Bottom -> fprintf fmt "BOTTOM\n"
        | Abs m -> fprintf fmt "%a\n" (Mat.print (Space.print)) m
    end;
    fprintf fmt "------------------------------------------\n"


  (* Dimensions *)

  let copy av =
    match !av with
      | Bottom -> ref Bottom
      | Abs m -> ref (Abs (Mat.copy m))

  let dimensions av =
    match !av with
      | Bottom ->
          raise (Invalid_argument "no sense to ask for variables involved into an empty DBM")
      | Abs m ->
          Var.SetteId.elements (Var.SetteId.remove Zero (Mat.indexes m))

  let dimension_add av id =
    match !av with
      | Bottom ->
          raise (Invalid_argument "no sense to add a dimension on an empty DBM")
      | Abs m ->
          let m = Mat.add m id Pinfty in
          Mat.set m id id Space.zero ; (* zero on diagonal *)
          ref (Abs m)

  (* Add dimension if not present *)
  let dimension_use av id =
    match !av with
      | Bottom ->
          raise (Invalid_argument "no sense to add a dimension on an empty DBM")
      | Abs m ->
          try (dimension_add av id) with (Invalid_argument _) -> av

  let dimension_remove av id =
    match !av with
      | Bottom ->
          raise (Invalid_argument "no sense to remove a dimension on an empty DBM")
      | Abs m ->
          let m' = (Mat.remove m id) in
            ref (Abs m')

  let dimension_rename av id_substitution =
    match !av with
      | Bottom ->
          raise (Invalid_argument "no sense to rename a dimension on an empty DBM")
      | Abs m ->
          let m' = (Mat.rename m id_substitution) in
            ref (Abs m')


  (* Complete lattices constants *)

  let bottom () =
    ref Bottom

  let top () =
    let m = Mat.empty () in
    let m_with_special_identificator =
      Mat.add m Zero Space.zero
    in
      ref (Abs m_with_special_identificator)

  let is_bottom av =
    match !av with
      | Bottom -> true
      | _ -> false


  (* pack variables management *) (* TO FACTOR *)

  let contract_packvariable m pv =
    Var.SetteId.fold
      (fun v m' ->
         (Mat.remove m' v)
      )
      pv
      m

  let expand_packvariable m pv =
    Var.SetteId.fold
      (fun v m' ->
         let m'' = (Mat.add m' v Pinfty) in
           Mat.set m'' v v Space.zero ; m''
      )
      pv
      m

  let contract_each_to_common_packvariable (m1, m2) =
    let vars1 = Mat.indexes m1 in
    let vars2 = Mat.indexes m2 in
    let intersection = Var.SetteId.inter vars2 vars1 in
    let diff1 = Var.SetteId.diff vars1 intersection in
    let diff2 = Var.SetteId.diff vars2 intersection in
    let m1' = contract_packvariable m1 diff1 in
    let m2' = contract_packvariable m2 diff2 in
      (m1', m2')

  let expand_each_to_same_packvariable (m1, m2) =
    let vars1 = Mat.indexes m1 in
    let vars2 = Mat.indexes m2 in
    let diff21 = Var.SetteId.diff vars2 vars1 in
    let diff12 = Var.SetteId.diff vars1 vars2 in
    let m1' = expand_packvariable m1 diff21 in
    let m2' = expand_packvariable m2 diff12 in
      (m1', m2')


  (* Lattice order and operators *)

  let canonical av =
    match !av with
      | Bottom -> ()
      | Abs m ->
          Graph.floyd_warshall m ;
          let consistent =
            Mat.diagonal_for_all (Space.is_leq Space.zero) m
          in
            av :=
              if (not (consistent))
              then Bottom
              else (Abs m)

  let is_leq av1 av2 = match !av1, !av2 with
    | Bottom, _ -> true
    | _, Bottom -> false
    | Abs m1, Abs m2 ->
        let m1', m2' =
          expand_each_to_same_packvariable (m1, m2)
        in
          Mat.for_all2 Space.is_leq m1' m2'

  let join av1 av2 = match !av1, !av2 with
    | Bottom, _ -> av2
    | _, Bottom -> av1
    | Abs m1, Abs m2 ->
        let m1', m2' =
          contract_each_to_common_packvariable (m1, m2)
        in
          ref (Abs (Mat.map2 Space.max m1' m2'))

  let join_list av_l =
    List.fold_left (fun res av -> join res av) (bottom ()) av_l

  let meet av1 av2 = match !av1, !av2 with
    | Bottom, _ -> av1
    | _, Bottom -> av2
    | Abs m1, Abs m2 ->
        let m1', m2' =
          expand_each_to_same_packvariable (m1, m2)
        in
        let av' =
          ref (Abs (Mat.map2 Space.min m1' m2'))
        in
           canonical av' ; av'

  let meet_list av_l =
    List.fold_left (fun res av -> meet res av) (top ()) av_l

  (* widening *)

  let widening av1 av2 = match !av1, !av2 with
    | Bottom, _ -> av2
    | _, Bottom -> raise (Invalid_argument "second operand of widening is empty although first one is not")
    | Abs m1, Abs m2 ->
        let m1', m2' =
          expand_each_to_same_packvariable (m1, m2)
        in
        let f = fun e1 e2 ->
          if Space.is_l e1 e2
          then Pinfty (* TODO : widening with landmarks (as ordered set) *)
          else e1
        in
          ref (Abs (Mat.map2 f m1' m2'))

  (* Abstract domain *)
  module Semantics =
  struct
    open Expr

    module Expression = LinearExpression (BaseIdentificator)
    module Guard = Expression.Guard
    module Command = Expression.Command

    open Expression
    open Guard
    open Command

    (* semantics operators *)

    (* Add the information  id1 >= id2 - c  to the abstract value *)
    let apply_potential_inequality av id1 id2 c =
      let av = dimension_use av id1 in
      let av = dimension_use av id2 in
      match !av with
       | Bottom -> raise (Invalid_argument "Abstract value should not be bottom here")
       | Abs m ->
           if not (Space.is_leq (Mat.get m id2 id1) (V c)) then
             Mat.set m id2 id1 (V c) ;
             canonical av
      ;
      av

    (* Add the information  id1 = id2 - c  to the abstract value *)
    let apply_potential_equality av id1 id2 c =
      let av = apply_potential_inequality av id1 id2 c
      in apply_potential_inequality av id2 id1 (-c)

    (* Forget the information about a variable *)
    let forget_variable av id =
      try
        match !av with
        | Bottom -> raise (Invalid_argument "no sense to forget a variable on an empty DBM")
        | Abs m ->
            if Var.SetteId.mem id (Mat.indexes m) then
              (Mat.project m id Pinfty ;
              Mat.set m id id (V 0))
      with Invalid_argument _ -> ()

    (* Add the information provided by an affectation  id1 <- id2 - c *)
    let apply_assignment av id1 id2 c =
      let av = dimension_add av Command in
      let av = apply_potential_equality av Command id2 c in
      forget_variable av id1 ;
      let av = apply_potential_equality av Command id1 0 in
      dimension_remove av Command

    (* Combine two semantic result from "and" or "or" expressions *)
    let combine_semantic_result result1 result2 = match result1, result2 with
      | Skip, _
      | _, Skip -> Skip
      | Abstracted, _
      | _, Abstracted -> Abstracted
      | Exact, Exact -> Exact

    (* Add the information provided by the guard *)
    let rec apply_guard av guard =
      let av = copy av in
      match !av with
        | Bottom -> (av, Exact)
        | Abs m ->
            match guard with
              | True -> (av, Exact)
              | False -> (bottom (), Exact)
              | And (e1, e2) ->
                  (* TODO : apply the first then the second, instead of doing meet operation ? *)
                  let av1, result1 = apply_guard av e1 in
                  let av2, result2 = apply_guard av e2 in
                  (meet av1 av2, combine_semantic_result result1 result2)
              | Or (e1, e2) ->
                  let av1, result1 = apply_guard av e1 in
                  let av2, result2 = apply_guard av e2 in
                  (join av1 av2, combine_semantic_result result1 result2)
              | Comparison (NotNul _) -> (av, Abstracted)
              | Comparison c ->
                  begin
                    (* Choose between apply_equality and apply_potential_inequality *)
                    let apply, t = match c with
                      | Nul t -> apply_potential_equality av, t
                      | NotNul t -> assert false
                      | PositiveNul t -> apply_potential_inequality av, t
                      | Positive t -> (fun id1 id2 c -> apply_potential_inequality av id1 id2 (c-1)), t
                    (* Apply the choosen operator *)
                    in match t with
                      (* s *)
                      | Value (Variable s) ->
                          (apply (Var.from_base s) Zero 0, Exact)
                      (* s + c *)
                      | Sum ([(Integer 1, Value (Variable s))], Integer c) ->
                          (apply (Var.from_base s) Zero c, Exact)
                      (* -s + c*)
                      | Sum ([(Integer -1, Value (Variable s))], Integer c) ->
                          (apply Zero (Var.from_base s) c, Exact)
                      (* s1 - s2 + c *)
                      | Sum ([(Integer 1, Value (Variable s1)) ; (Integer -1, Value (Variable s2))], Integer c)
                      | Sum ([(Integer -1, Value (Variable s2)) ; (Integer 1, Value (Variable s1))], Integer c) ->
                          (apply (Var.from_base s1) (Var.from_base s2) c, Exact)
                      | _ -> (av, Abstracted)
                  end
              | _ -> raise (NotImplemented "guard not admitted by the analysis")


    (* Apply a command to the abstract value *)
    (* TODO : give a real semantic result *)
    let apply_command av cmd =
      let av = copy av in
      match !av with
        | Bottom -> (av, Exact)
        | Abs m ->
            match cmd with
              | StdExpression.NumericalAssign (Variable s1, expression) ->
                begin
                  let id1 = Var.from_base s1 in
                  match expression with
                    (* id1 = c *)
                    | Value (Const (Integer c)) ->
                        (apply_assignment av id1 Zero (-c), Exact)
                    (* id1 = id2 *)
                    | Value (Variable s2) ->
                        (apply_assignment av id1 (Var.from_base s2) 0, Exact)
                    (* id1 = id2 + c *)
                    | Sum ([(Integer 1, Value (Variable s2))], Integer c) ->
                        (apply_assignment av id1 (Var.from_base s2) (-c), Exact)
                    (* id1 = Unknown *)
                    | _ ->
                        (forget_variable av id1 ; av, Abstracted)
                end
              |  StdExpression.NumericalAssign (_, expression) -> (av, Skip)
              | _ -> raise (NotImplemented "command not admitted for assignment by the analysis")
  end

  let print fmt av =
    match !av with
      | Bottom -> fprintf fmt "BOTTOM"
      | Abs m ->
          let print_string_list fmt = function
            | [] -> ()
            | s :: l ->
                fprintf fmt "%s" s ;
                List.iter (fun s -> fprintf fmt ",@ %s" s) l
          in
          let print_constant_term fmt = function
            | c when c > 0 -> fprintf fmt " + %a" pp_print_int c
            | c when c < 0 -> fprintf fmt " - %a" pp_print_int (-c)
            | c -> ()
          in
            let coeff_to_string id1 id2 e l =
              match e with
                | Pinfty -> l
                | Ninfty -> l
                | V c ->
                    try
                      let comparison p1 v1 p2 v2 c =
                        if Space.is_eq (Space.neg e) (Mat.get m id2 id1) then
                          (if Var.OrderedId.compare id1 id2 > 0 then raise Exit ;
                          fprintf str_formatter "%a = %a%a" p1 v1 p2 v2 print_constant_term (-c))
                        else if c <> 0 then
                          fprintf str_formatter "%a > %a%a" p1 v1 p2 v2 print_constant_term (-c-1)
                        else
                          fprintf str_formatter "%a >= %a%a" p1 v1 p2 v2 print_constant_term (-c)
                        ;
                        flush_str_formatter ()
                      in
                      match id1, id2 with
                        | Command, _
                        | _, Command -> raise (Invalid_argument "Command special identificator should not be here")
                        | id1, id2 when id1 = id2 -> l
                        (* 0 + c >= id1 *)
                        | Normal _, Zero -> comparison pp_print_int c Var.print id1 0 :: l
                        (* id2 + c >= 0 *)
                        | Zero, Normal _ -> comparison Var.print id2 pp_print_int (-c) 0 :: l
                        (* id2 + c >= id1 *)
                        | Normal _, Normal _ -> comparison Var.print id2 Var.print id1 c :: l
                        | id1, id2 -> comparison Var.print id2 Var.print id1 c :: l
                    with
                      | Exit -> l
            in
            let string_list =
              Mat.foldi coeff_to_string m []
            in
              print_string_list fmt string_list
end


(** DBM with standard identificators *)

module Zdbm = ZDbmLattice(StdIdentificator)

