
open Types
open Identificator

module type VectorSpace =
sig
  type v
  type s

  val plus : v -> v -> v
  val minus : v -> v -> v
  val invers : v -> v
  val mul : s -> v -> v
  val div : s -> v -> v

  val zero : v
  val one : s
end

module type Matrix = sig
  (* type of indexes *)
  type row_index
  type column_index
  (* matrice type *)
  type 'a t

  (**
   *  Constructors
   *)
  (* Creates an empty matrix *)
  val empty : unit -> 'a t
  (* Returns a copy of the matrix *)
  val copy : 'a t -> 'a t

  (**
   *  Indexes
   *)
  (* Add a row or a column *)
  val add_row : 'a t -> 'a -> 'a t
  val add_column : 'a t -> 'a -> 'a t
  (* Insert a row or a column at specified index *)
  val insert_row : 'a t -> row_index -> 'a -> 'a t
  val insert_column : 'a t -> column_index -> 'a -> 'a t
  (* Removes a row or a column at specified index *)
  val remove_row : 'a t -> row_index -> 'a t
  val remove_column : 'a t -> column_index -> 'a t
  (* Moves a row or a column at specified index *)
  (* Semantic of a move is implementation dependant *)
  val move_row : 'a t -> row_index -> row_index -> unit
  val move_column : 'a t -> row_index -> row_index -> unit
  (* Get the matrix row count *)
  val row_count : 'a t -> int
  (* Get the matrix column count *)
  val column_count : 'a t -> int

  (**
   *  Basic operations
   *)
  (* Get matrix coefficient at specified index *)
  val get : 'a t -> row_index -> column_index -> 'a
  (* Set matrix coefficient at specified index *)
  val set : 'a t -> row_index -> column_index -> 'a -> unit
  (* Change coefficient *)
  val change : 'a t -> row_index -> column_index -> ('a -> 'a) -> unit

  (**
   *  Iterators
   *)
  (* Fold the given row *)
  val row_fold : row_index -> ('a -> 'b -> 'b) -> 'a t -> 'b -> 'b
  val row_foldi : row_index -> (column_index -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
  (* Fold the whole matrix *)
  val cell_fold : ('a -> 'b -> 'b) -> 'a t -> 'b -> 'b
  val cell_foldi : (row_index -> column_index -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
  (* Fold each row then fold the folding *)
  val fold : ('a -> 'b -> 'b) -> ('b -> 'c -> 'c) -> 'a t -> 'b -> 'c -> 'c
  val foldi : (row_index -> column_index -> 'a -> 'b -> 'b) -> (row_index -> 'b -> 'c -> 'c) -> 'a t -> 'b -> 'c -> 'c
  (* Iter *)
  val iter : ('a -> unit) -> 'a t -> unit
  val iteri : (row_index -> column_index -> 'a -> unit) -> 'a t -> unit
  (* Modification in place *)
  val modify : ('a -> 'a) -> 'a t -> unit
  val modifyi : (row_index -> column_index -> 'a -> 'a) -> 'a t -> unit
  (* Mapping *)
  val map : ('a -> 'b) -> 'a t -> 'b t
  val map2 : ('a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t
  val mapi : (row_index -> column_index -> 'a -> 'b) -> 'a t -> 'b t
  val mapi2 : (row_index -> column_index -> 'a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t
  (* Check if condition hold on one coefficient *)
  val exists : ('a -> bool) -> ('a t) -> bool
  val exists2 : ('a -> 'a -> bool) -> ('a t) -> ('a t) -> bool
  (* Check if condition hold on each coefficient *)
  val for_all : ('a -> bool) -> ('a t) -> bool
  val for_all2 : ('a -> 'a -> bool) -> ('a t) -> ('a t) -> bool


  (**
   *  Matrix display
   *)
  val print : (Format.formatter -> 'a -> unit) -> Format.formatter -> 'a t -> unit

  (**
   *  Linear Algebra
   *)
  module LinearAlgebra (Space : VectorSpace) :
  sig
    type v = Space.v
    type s = Space.s

    (* Add the first row multiplied by the first coefficient to the second row
      multiplied by the second coefficient *)
    val combine_rows : v t -> s -> row_index -> s -> row_index -> unit
    (* Exchanges two rows *)
    val exchange_rows : v t -> row_index -> row_index -> unit
    (* Multiplies the given row *)
    val multiply_row : v t -> s -> row_index -> unit
    (* Divide the given row *)
    val divide_row : v t -> s -> row_index -> unit
  end
end


(**
   Standard Square Matrices

   implemented as array of arrays
*)

module StandardMatrix : Matrix with
  type row_index = int and
  type column_index = int
