
open Format

open Utils
open Matrix
open Identificator


(* Notes : Normalisations are made systematically to ensure that no
 * linear union are done on non canonical matrices. Design must be
 * reviewed. -- suppression temporaire
 *
 * Variable add/remove is done here. This is not a good thing, but
 * much easyier as doing it in Multiset.
 *)


(*----------------------------------------------------------------------*
 * Signatures                                                           *
 *----------------------------------------------------------------------*)

module type LinearSystem = sig
  type variable
  type t

  (* Creates an empty system. *)
  val empty : unit -> t
  (* Returns a copy of the system. *)
  val copy : t -> t
  (* Add a variable to the system with no information about it. *)
  val add_variable : t -> variable -> unit
  (* Remove a variable from the system. At most one equation will be removed to preserve correction. *)
  val remove_variable : t -> variable -> unit
  (* Rename a variable in the system. *)
  val rename_variable : t -> variable -> variable -> unit
  (* Rename simultaneously a list of variable in the system. *)
  val rename_list : t -> (variable * variable) list -> unit
  (* Forget all information known about the variable. At most one equation will be removed. *)
  val forget_variable : t -> variable -> unit
  (* Tell if a variable exists in the system *)
  val mem : variable -> t -> bool
  (* List of variables *)
  val variables : t -> variable list
  (* Add an equation variable1 = variable2 into the system. *)
  val add_equality : t -> variable -> variable -> unit
  (* Add an equation Sum a_i v_i = 0 into the system. *)
  val add_equation : t -> (variable * int) list -> unit
  (* Substitute a variable by a list of variable. *)
  val substitute : t -> variable -> (variable * int) list -> unit
  (* Get a list of equations *)
  val equations : t -> (variable * variable) list
  (* Tell if x = y *)
  val equal_in_system : t -> variable -> variable -> bool
  (* Normalize the system in place. *)
  val normalize : t -> unit
  (* Linear union between two systems. Does not modifiy system1 and system2 (work on copies) *)
  val linear_union : t -> t -> t
  (** Inclusion test. *)
  val is_leq : t -> t -> bool
  (** Folding *)
  val foldi : (variable -> int -> 'a -> 'a) -> ('a -> 'b -> 'b) -> t -> 'a -> 'b -> 'b
  (** System display. *)
  val print : Format.formatter -> t -> unit
end


(*----------------------------------------------------------------------*
 * Standard implementation                                              *
 *----------------------------------------------------------------------*)

module IntegerSpace =
struct
  type v = int
  type s = int

  let plus = (+)
  let minus = (-)
  let invers = fun x -> -x
  let mul = ( * )
  let div = fun x y -> y / x

  let zero = 0
  let one = 1
end

module StandardLinearSystem (Variables: Identificator) :
  LinearSystem with type variable = Variables.t
 = struct

  module Implementation = StandardMatrix
  module Mappe = Variables.MappeId
  module Sette = Variables.MappeId.Setkey
  module LinearAlgebra = Implementation.LinearAlgebra (IntegerSpace)

  type variable = Variables.t


  type t =
    {
      mutable matrix: int Implementation.t ;
      mutable idf2ind: int Mappe.t; (* TODO : hash *)
      mutable ind2idf: variable array;
    }

  let empty () =
    {
      matrix = Implementation.empty () ;
      idf2ind = Mappe.empty ;
      ind2idf = [||]
    }

  (* Returns a copy of the system. *)
  let copy system =
    {
      system with
      matrix = Implementation.copy system.matrix ;
      ind2idf = Array.copy system.ind2idf
    }

  (*
   * Frequently used implementation functions
   *)

  let row_count system = Implementation.row_count system.matrix
  let column_count system = Implementation.column_count system.matrix
  let get system i j = Implementation.get system.matrix i j
  let set system = Implementation.set system.matrix
  let add_row system =
    let indice = row_count system in
    begin
      system.matrix <- Implementation.insert_row system.matrix indice 0 ;
      indice
    end
  let remove_row system index = system.matrix <- Implementation.remove_row system.matrix index


  (*
   * Index conversion
   *)

  let ind : t -> variable -> int = fun system variable ->  Mappe.find variable system.idf2ind
  let idf : t -> int -> variable = fun system indice -> system.ind2idf.(indice)


  (*
   * System display
   *)
  let print fmt system =
    array_print (Variables.print) fmt system.ind2idf ~first:"[|\t" ~sep:"\t" ~last:"\t|]\n" ;
    Implementation.print (fun fmt x -> Format.fprintf fmt "%d" x) fmt system.matrix


  (*
   * Integer operator definition for linear algebra
   *)

  let multiply_row system c i =
    if c <> 1 then
      LinearAlgebra.multiply_row system.matrix c i

  let divide_row system c i =
    LinearAlgebra.divide_row system.matrix c i

  (* Add the first row multiplied by the first coefficient to the second row
     multiplied by the second coefficient *)
  let combine_rows system c1 i1 c2 i2 =
    if c1 = 0 then
      multiply_row system c2 i2
    else
      LinearAlgebra.combine_rows system.matrix c1 i1 c2 i2

  let exchange_rows system i1 i2 =
    LinearAlgebra.exchange_rows system.matrix i1 i2


  (*
   * Variable manipulation
   *)

  let mem variable system = Mappe.mem variable system.idf2ind

  let variables system = Array.to_list system.ind2idf


  (* Add variable without verification *)
(*
  let add_variable' system variable =
    let indice = column_count system in
      system.idf2ind <- Mappe.add variable indice system.idf2ind ;
      system.ind2idf <- Array.append system.ind2idf [|variable|] ;
      system.matrix <- Implementation.add_column system.matrix indice 0 ;
      indice
*)

  let position_in_sette sette variable =
    let indices_list = Sette.elements (sette)
    in
      let i = ref (-1) in
        List.fold_left
          (fun x v -> i := !i + 1 ; if v = variable then !i else x)
          (-1)
          indices_list

  let rebuild_association system =
    let i = ref (-1) in
    Array.fold_left
      (fun mappe variable -> i := !i + 1 ; Mappe.add variable !i mappe)
      Mappe.empty
      system.ind2idf

  let add_variable' system variable =
    let indice = position_in_sette (Sette.add variable (Mappe.maptoset system.idf2ind)) variable in
    system.ind2idf <- array_insert system.ind2idf indice variable ;
    system.idf2ind <- rebuild_association system ;
    system.matrix <- Implementation.insert_column system.matrix indice 0 ;
 (*
      Mappe.print (Variables.print) (fun fmt x -> Format.fprintf fmt "%d" x) std_formatter system.idf2ind ~first:"[|\t" ~sep:"\t" ~last:"\t|]\n" ;
      array_print (Variables.print) std_formatter system.ind2idf ~first:"[| " ~sep:" " ~last:" |]\n" ;
      Implementation.print (fun fmt x -> Format.fprintf fmt "%d" x) std_formatter system.matrix ;
*)
      indice

  let add_variable system variable =
    if Mappe.mem variable system.idf2ind then
      raise (Invalid_argument "attempt to add to a linear system a variable already present")
    else
      ignore (add_variable' system variable)

  let add_variable_ifndef system variable =
    try Mappe.find variable system.idf2ind
    with Not_found -> add_variable' system variable

  let forget_variable system variable =
    let n = row_count system in
    let j = add_variable_ifndef system variable in
    let l = ref 0 in (* l is the last coefficient <> 0 on the column *)
    begin
    try
      for i = n - 1 downto 0 do
        if get system i j <> 0 then
        begin
          l := i ;
          raise Exit
        end
      done
    with Exit -> (* some non nul coefficient found *)
      for i = 0 to !l - 1 do
        combine_rows system (-(get system i j)) !l (get system !l j) i
      done ;
      remove_row system !l
    end

  let remove_variable system variable =
    forget_variable system variable ;
    if (not (Mappe.mem variable system.idf2ind)) then
      raise (Invalid_argument "attempt to remove from a linear system an unknown variable")
    else
      let indice = Mappe.find variable system.idf2ind in
      system.ind2idf <- array_remove system.ind2idf indice ;
      system.idf2ind <- rebuild_association system ;
      system.matrix <- Implementation.remove_column system.matrix indice

  (* After renaming, system is not canonical anymore ! *)
  let rename_variable system old_variable new_variable =
(*
    let indice = Mappe.find old_variable system.idf2ind in
    begin
      system.ind2idf.(indice) <- new_variable ;
      system.idf2ind <- Mappe.add new_variable indice (Mappe.remove old_variable system.idf2ind)
    end
*)
    let i1 = position_in_sette (Mappe.maptoset system.idf2ind) old_variable in
    let i2 = position_in_sette (Sette.add new_variable (Sette.remove old_variable (Mappe.maptoset system.idf2ind))) new_variable in
    array_move system.ind2idf i1 i2 ;
    system.ind2idf.(i2) <- new_variable ;
    system.idf2ind <- rebuild_association system ;
    Implementation.move_column system.matrix i1 i2

  let rename_list system l =
    let old_matrix = Implementation.copy system.matrix in
    let old_indexes = List.fold_left
      (fun m' (i,j) -> Mappe.add j (Mappe.find i system.idf2ind) m')
      (List.fold_left (fun m' (i,_j) -> Mappe.remove i m') system.idf2ind l)
      l
    in
    let traductor = Array.make (column_count system) 0 in
    let i = ref (-1) in
    Mappe.iter
      begin fun (name : variable) (oldindex : int) ->
        i := !i + 1 ;
        system.ind2idf.(!i) <- name ;
        traductor.(!i) <- oldindex
      end
      old_indexes
    ;
    system.idf2ind <- rebuild_association system ;
    Implementation.modifyi
      begin fun i j _ ->
        Implementation.get old_matrix i traductor.(j)
      end
      system.matrix


  (*
   * Iterations
   *)
  (** Folding *)
  let foldi f g system =
    Implementation.foldi
      (fun i j -> f (idf system j))
      (fun i -> g)
      system.matrix


  (*
   * System operations
   *)

  (* Divide each line by the pgcd of the line *)
  let normalize_rows system =
    let n = row_count system in
    let rec pgcd a = function 0 -> a | b -> pgcd b (a mod b) in
    for i = 0 to n - 1 do
      match Implementation.row_fold i pgcd system.matrix 0 with
      | 0 -> remove_row system i
      | a -> divide_row system (abs a) i
    done

  (* reduced row echelon form with Gauss Algorithm *)
  let normalize system =
    let n, m = row_count system, column_count system in
    let i = ref 0 in
    for j = 0 to m - 1 do
    (* Mappe.iter (fun _ j -> *)
      let l = ref 0 in (* l is the original row of the pivot *)
      try
        for k = !i to n - 1 do
          if get system k j <> 0 then
          begin
            l := k ;
            raise Exit
          end
        done
      with Exit -> (* Pivot found *)
        begin
          exchange_rows system !i !l ;
          for k = 0 to n - 1 do
            if k <> !i then
              combine_rows system (-(get system k j)) !i (get system !i j) k
          done ;
          i := !i + 1
        end
    (* ) system.idf2ind ; *)
    done ;
    normalize_rows system

  let add_equation system l =
    let variables = List.map fst l in
    (* add_variable_ifndef returns new index of variable if inserted but it can not be trusted since
       indexes may have change after many addition of variable. Indexes are hence recomputed after
       insertions *)
    List.iter (fun v -> ignore (add_variable_ifndef system v)) variables ;
    let indexes = List.map (ind system) variables in
    let coefficients = List.map snd l in
    let i = add_row system in
    List.iter2 (set system i) indexes coefficients ;
    normalize system

  let add_equality system variable1 variable2 =
    add_equation system [(variable1, 1) ; (variable2 , -1)]

  (* Get a list of equations *)
  let equations system =
     let n, m = row_count system, column_count system in
     let rec equations acc = function
     | i when i = n -> acc
     | i ->
         let vp1, vm1 = ref (-1), ref (-1) in
         try
           for j = 0 to m - 1 do
             match get system i j with
             | 0 -> ()
             | 1 -> if !vp1 = -1 then vp1 := j else raise Exit
             | -1 -> if !vm1 = -1 then vm1 := j else raise Exit
             | _ -> raise Exit
           done ;
           equations ((idf system !vp1, idf system !vm1) :: acc) (i+1)
         with exit -> equations acc (i+1)
     in
       equations [] 0

  (* Tell if x = y *)
  let equal_in_system system v1 v2 =
    let system' = copy system in
    add_equality system' v1 v2 ;
    row_count system' <= row_count system (* no information gained *)

  (* Substitute a variable by a list of variable. *)
  let substitute system variable replacement_list =
    let n = row_count system in
    let rec add_variables = function
    | [] -> ()
    | (variable, b) :: l -> ignore (add_variable_ifndef system variable) ; add_variables l
    in
    add_variables replacement_list ;
    let j = add_variable_ifndef system variable in
    let rec substitute_in_rows = function
    | [] -> ()
    | (variable, b) :: l ->
        let k = ind system variable in
        for i = 0 to n - 1 do
          let a = get system i j in
          set system i k ((get system i k) + a * b)
        done ;
        substitute_in_rows l
    in
      substitute_in_rows replacement_list ;
      normalize system


  let linear_union system1 system2 =
    (* (* Variable management *)
    if not (Mappe.Setkey.equal (Mappe.maptoset system1.idf2ind) (Mappe.maptoset system2.idf2ind)) then
      raise (Invalid_argument "attempt to make a linear union between two linear systems which do not have the same variables")
    else
    *)
    let system1, system2 = copy system1, copy system2 in
    (* Variable management *)
    let variable_set_1, variable_set_2 = Mappe.maptoset system1.idf2ind, Mappe.maptoset system2.idf2ind in
    let diff1, diff2 = Sette.diff variable_set_1 variable_set_2, Sette.diff variable_set_2 variable_set_1 in
    Sette.iter (fun variable -> ignore (add_variable' system1 variable)) diff2 ;
    Sette.iter (fun variable -> ignore (add_variable' system2 variable)) diff1 ;
    let m = column_count system1 in
    let i = ref 0 in
    for j = 0 to m - 1 do
    (* Mappe.iter (fun _ j -> *)
      let n1, n2 = row_count system1, row_count system2 in
      match
        (if !i < n1 then get system1 !i j else 0),
        (if !i < n2 then get system2 !i j else 0)
      with
        0, 0 ->
          (let l = ref 0 in (* l is the last coefficient <> 0 on the column *)
          try
            for k = !i - 1 downto 0 do
              if get system1 k j <> get system2 k j then
              begin
                l := k ;
                raise Exit
              end
            done
          with Exit -> (* some coefficient differ *)
          begin
            let dl = get system1 !l j - get system2 !l j in
            for k = 0 to !l - 1 do
              let dj = get system1 k j - get system2 k j in
              begin
                combine_rows system1 (-dj) !l dl k ;
                combine_rows system2 (-dj) !l dl k
              end
            done ;
            remove_row system1 !l ;
            remove_row system2 !l ;
            i := !i - 1
          end)
      | a, 0 ->
         for k = 0 to !i - 1 do
           combine_rows system1 (get system2 k j) !i a k ;
           multiply_row system2 a k
         done ;
         remove_row system1 !i
      | 0, b ->
         for k = 0 to !i - 1 do
           combine_rows system2 (get system1 k j) !i b k ;
           multiply_row system1 b k
         done ;
         remove_row system2 !i
      | a, b ->
          multiply_row system1 b !i ;
          multiply_row system2 a !i ;
          i := !i + 1
    (* ) system1.idf2ind ;*)
    done ;
    normalize_rows system1 ;
    system1

  (** Inclusion test *)
  let is_leq system1 system2 =
    let union = linear_union system1 system2 in
      row_count system2 = row_count union
end
