(* Enkidu
    Version 2.0
*)

open Format

open Types
open Utils

open Arg

(* TODO *)

let possible_analysis_l =
  [
    "reachability-dbm";
    "reachability-ms";
  ]

let get_analysis_arguments = function
  | "reachability-dbm" -> Analyzer.ClassicalReachability.ZonesReachability.Configuration.argument_specification
  | "reachability-ms" -> Analyzer.ClassicalReachability.MultisetReachability.Configuration.argument_specification
  | s -> raise (NotImplemented ("given analysis : " ^ s))

let get_analysis_default_configuration = function
  | "reachability-dbm" -> Analyzer.ClassicalReachability.ZonesReachability.Configuration.default_configuration
  | "reachability-ms" -> Analyzer.ClassicalReachability.MultisetReachability.Configuration.default_configuration
  | s -> raise (NotImplemented ("given analysis : " ^ s))

let get_analysis = function
  | "reachability-dbm" -> Analyzer.ClassicalReachability.ZonesReachability.run_and_print
  | "reachability-ms" -> Analyzer.ClassicalReachability.MultisetReachability.run_and_print
  | s -> raise (NotImplemented ("given analysis : " ^ s))


(* types *)

(* arguments retrieval *)

let program_name =  Sys.argv.(0)

let default_usage_message = "Usage: " ^ program_name ^ " [options] file \n\nwhere options are : "

let specialized_usage_message analysis = "Usage: " ^ program_name ^ " [options] file \n\nwhere options (for analysis `" ^ analysis ^ "') are : "

let default_configuration =
  [
    ("analysis", Configuration.String "reachability-ms") ;
    ("verbose", Configuration.Int 0) ;
    ("dot_output", Configuration.Bool false)
  ]

let possible_analyses =
  List.fold_left
    (fun possible_analyses s ->
       possible_analyses ^ "|" ^ s
    )
    (List.hd possible_analysis_l)
    (List.tl possible_analysis_l)

let set_file configuration file =
  if Sys.file_exists file
  then
    Configuration.set_string configuration "file" file
  else
    raise (Arg.Bad ("file " ^ file ^ " does not exist"))

let set_analysis configuration analysis =
  if List.mem analysis possible_analysis_l
  then
    Configuration.set_string configuration "analysis" analysis
  else
    raise (Arg.Bad ("analysis " ^ analysis ^ " unknown"))

let set_verbose configuration level =
  if (level >= 0 && level <= 3)
  then
    Configuration.set_int configuration "verbose" level
  else
    raise (Arg.Bad ("verbose mode " ^ (string_of_int level) ^ " unknown"))

let default_arguments configuration =
  [("-analysis",
    Arg.String (set_analysis configuration),
    "A  Run analysis A on the given program (A = " ^ possible_analyses ^ ")");
   ("-verbose",
    Arg.Int (set_verbose configuration),
    "n  Verbose level (n in [0..3])");
   ("-version",
    Arg.Unit (fun () -> raise(Version)),
    " Print version and exit")
 ]


let get_arguments () =
  let configuration = Configuration.init default_configuration in
  let usage_message = ref default_usage_message in
  let arguments = ref (default_arguments configuration) in
  try
  (
    (* Get the analysis *)
    begin
      try
        let index = array_find "-analysis" Sys.argv in
        if index + 1 < Array.length Sys.argv then
          set_analysis configuration (Sys.argv.(index + 1))
        else
          raise (Arg.Bad (program_name ^ ": option `-analysis' needs an argument."))
      with Not_found -> ()
    end
    ;
    let analysis = Configuration.get_string configuration "analysis" in
    (* Add specifications *)
    Configuration.append configuration (get_analysis_default_configuration analysis) ;
    usage_message := specialized_usage_message analysis ;
    arguments := get_analysis_arguments analysis configuration @ !arguments ;
    (* Parse the command line *)
    Arg.parse !arguments (set_file configuration) !usage_message ;
    if not (Configuration.isset configuration "file") then
      raise (Arg.Bad (program_name ^ ": no file given"))
    ;
    configuration
  )
  with
    | Arg.Bad s ->
        eprintf "%s@." s ;
        Arg.usage !arguments !usage_message ;
        raise Usage
    | Version ->
        printf "Version: 2.0 / July 2009@." ;
        raise Version


let _ =
  try
    (* Get configuration *)
    let configuration = get_arguments () in
    let file_name = Configuration.get_string configuration "file" in
    let analysis_name = Configuration.get_string configuration "analysis"
    in
    let module AnalysisFor (C : Compiler.Compiler) =
      struct
        let run () =
          (* Get graph *)
          let program_graph = C.compile_file file_name in
          (* Get and run analysis *)
          let run_and_print = get_analysis analysis_name in
          run_and_print configuration program_graph C.print_state_data C.print_transition_data C.print_program_data std_formatter
      end
    in
    (* Find the apropriate language *)
    if Filename.check_suffix file_name ".fst" then
      let module A = AnalysisFor (Analyzer.Languages.FastLanguage.FastCompiler.Compiler) in
      A.run ()
    else if Filename.check_suffix file_name ".c" then
      let module A = AnalysisFor (Analyzer.Languages.CLanguage.CCompiler.Compiler) in
      A.run ()
    else
      raise FileFormat
  with
    | Version ->
        exit(0)
    | Usage ->
        exit(2)
    | FileFormat ->
        printf "File format not recognized\n" ;
        exit(3)

