
(* Internal exceptions *)

exception FoundInt of int


(* --- Tuple Utils ---*)

let lexicographic_compare =
  fun (f1, f2) (a1, a2) (b1, b2) ->
  let r = f1 a1 b1 in
    if r = 0 then
      f2 a2 b2
    else
      r


(* --- Array Utils ---*)

(* Find the index of ab element *)
let array_find e a =
  try
    Array.iteri (fun i e' -> if e = e' then raise (FoundInt i)) a ;
    raise Not_found
  with FoundInt i ->
    i

(* Remove the element at index i from the array *)
let array_remove (a : 'a array) (i : int) : 'a array =
  Array.concat [Array.sub a 0 i ; Array.sub a (i + 1) (Array.length a - i - 1)]

(* Insert the element e at index i *)
let array_insert (a : 'a array) (i : int) (x : 'a) : 'a array =
  Array.concat [Array.sub a 0 i ; [|x|] ; Array.sub a i (Array.length a - i)]

(* Add the element e at the end of the array *)
let array_add (a : 'a array) (x : 'a) : 'a array =
  Array.append a [|x|]

(* Move the element e from index i1 to index i2 *)
let array_move (a : 'a array) (i1 : int) (i2 : int) : unit =
  let x = a.(i1) in
  if i1 < i2 then
    Array.blit a (i1+1) a i1 (i2 - i1)
  else
    Array.blit a i2 a (i2+1) (i1 - i2)
  ;
  a.(i2) <- x

(* Print the array *)
let array_print
  ?(first : (unit, Format.formatter, unit) format = ("[| " : (unit, Format.formatter, unit) format))
  ?(sep : (unit, Format.formatter, unit) format = (" ; ":(unit, Format.formatter, unit) format))
  ?(last : (unit, Format.formatter, unit) format = (" |]":(unit, Format.formatter, unit) format))
  (print_element:Format.formatter -> 'a -> unit)
  (formatter:Format.formatter)
  (a:'a array)
  : unit
  =
  Format.fprintf formatter first;
  let firstelt = ref true in
  Array.iter
    (begin fun elt ->
      if !firstelt then firstelt := false else Format.fprintf formatter sep;
      Format.fprintf formatter "%a" print_element elt
    end)
    a;
  Format.fprintf formatter last


(* --- List Utils ---*)

let rec list_compare compare l1 l2 = match l1, l2 with
  | [], [] -> 0
  | _, [] -> 1
  | [], _ -> -1
  | a :: l1, b :: l2 ->
      let c = compare a b in
      if c = 0 then list_compare compare l1 l2 else c

let list_print
  ?(first : (unit, Format.formatter, unit) format = ("[ " : (unit, Format.formatter, unit) format))
  ?(sep : (unit, Format.formatter, unit) format = (" ;@, ":(unit, Format.formatter, unit) format))
  ?(last : (unit, Format.formatter, unit) format = (" ]":(unit, Format.formatter, unit) format))
  (print_element:Format.formatter -> 'a -> unit)
  (formatter:Format.formatter)
  (list:'a list)
  : unit
  =
(*
  match list with
    | [] -> Format.fprintf formatter "%{%}%{%}" first last
    | x :: l ->
        Format.fprintf formatter "%{%}%a%a%{%}"
          first
          print_element x
          (fun fmt -> List.iter (Format.fprintf fmt "%{%}%a" sep print_element)) l
          last
*)
  match list with
    | [] -> Format.fprintf formatter "[]"
    | x :: l ->
        Format.fprintf formatter "[%a%a]"
          print_element x
          (fun fmt -> List.iter (Format.fprintf fmt " ; %a" print_element)) l

let rec sorted_list_unique = function
  | x1 :: x2 :: l when x1 = x2 -> sorted_list_unique (x1 :: l)
  | x1 :: x2 :: l -> x1 :: sorted_list_unique (x2 :: l)
  | l -> l


(* --- Configuration --- *)

module Configuration =
struct
  type key = string
  type value = String of string | Bool of bool | Int of int
  type values = (key * value) list
  exception WrongType of key * value

  type t =
    {
      mutable values : (key * value) list
    }


  let empty () =
    {
      values=[]
    }

  let init value_list =
    {
      values=value_list
    }


  let append configuration value_list =
    configuration.values <- configuration.values @ value_list


  let isset configuration key =
    List.mem_assoc key configuration.values

  let unset configuration key =
    configuration.values <- List.remove_assoc key configuration.values


  let set configuration key value =
    unset configuration key ;
    configuration.values <-(key, value) :: configuration.values

  let set_string configuration key value =
    set configuration key (String value)

  let set_int configuration key value =
    set configuration key (Int value)

  let set_bool configuration key value =
    set configuration key (Bool value)

  let set_flag configuration key () =
    set configuration key (Bool true)

  let reset_flag configuration key () =
    set configuration key (Bool false)


  let get configuration key =
    List.assoc key configuration.values

  let get_string configuration key =
    match get configuration key with
      | String value -> value
      | v -> raise (WrongType (key, v))

  let get_int configuration key =
    match get configuration key with
      | Int value -> value
      | v -> raise (WrongType (key, v))

  let get_bool configuration key =
    match get configuration key with
      | Bool value -> value
      | v -> raise (WrongType (key, v))

  let get_flag configuration key =
    isset configuration key &&
    get_bool configuration key

end

(* Labeled Transition Systems *)

module LabeledTransitionSystem =
struct

  module TSymbol =
  struct
    type vertex = Symbol.t
    type hedge = Symbol.t

    let vertex_dummy = Symbol.add "vertex dummy"
    let hedge_dummy = Symbol.add "hedge dummy"

    module SetV = Symbol.Set
    module SetH = Symbol.Set
    module HashV = Hashhe.Make(Symbol.HashedType)
    module HashH = Hashhe.Make(Symbol.HashedType)
  end

  module Graph = SHGraph.Make(TSymbol)

  type ('a,'b,'c) t = {
    graph: ('a,'b,'c) Graph.t ;
    init_states: TSymbol.SetV.t
  }

  let print_dot
    (state_data_printer : Format.formatter -> 'a -> unit)
    (transition_data_printer : Format.formatter -> 'b -> unit)
    (program_data_printer : Format.formatter -> 'c -> unit)
    (fmt : Format.formatter)
    ({graph=graph} : ('a,'b,'c) t) : unit
  =
    let title =
      Format.fprintf Format.str_formatter "%a" program_data_printer (Graph.info graph) ;
      Format.flush_str_formatter ()
    in
    Graph.print_dot
      ~title:title
      Symbol.print
      Symbol.print
      (fun fmt _ s -> state_data_printer fmt s)
      (fun fmt _ e -> transition_data_printer fmt e)
      fmt
      graph

end
